/****************************************************************************
*                                                                           *
*              / SCHEDDSP.PRG                                               *
*                                                                           *
*         Autor: M.Rinninger                                                *
*   Erstellt am:  01-11-95      /  08:49:22am                               *
*  berarbeitet: 02-25-95      / 09:28:35am                               *
*           von:                                                            *
*     Copyright: (c) 1995 by Ob-Tec Austria                                 *
*                ontick() is (c) by Ted Means and                           *
*                is slighly modified by the aithor                          *
****************************************************************************/

   #include "schedule.ch"
   #include "class(y).ch"

   #translate .receiver    => \[ 1]
   #translate .message     => \[ 2]
   #translate .param       => \[ 3]
   #translate .delay       => \[ 4]
   #translate .date        => \[ 4]
   #translate .timer       => \[ 5]
   #translate .reinstall   => \[ 6]
   #translate .pid         => \[ 7]
   #define ENTRY_SIZE 7

   #define QUEUE_SIZE 64

   static _nDelay      := 0
   static _isActive    := false
   static _isChecking  := false
   static _toDo        := {}
   static _queue       := {}
   static _read        := 0
   static _write       := 0

   static _eval
   static _pid

function SchedDsp();return( nil)

/* Function Ŀ
           Name: SchedAdd()            Docs:                                
    Description: create a new entry in the scheduler process queue.         
                 Defaults some parameters to recomended values              
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 04:46:31pm            Time updated: 04:46:31pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: receiver  - either a codeblock or a object                 
               : symbol    - for a codeblock, this is always 'eval()',      
                             otherwise a method                             
               : parameter - this single value is transferd to the symbol   
                             at execution. Use an array if you need more    
                             than one parameter.                            
               : nDelay    - numbers of ticks to wait until execution       
               : lInstall  - if true, process will be reinserted into the   
                             process queue automaticaly.                    
   Return Value: _pid      - process id, could be used to kill a process.   
       See Also: SchedKill(), SchedRemove()                                 
   */
function SchedAdd( receiver, symbol, parameter, nDelay, lInstall)
//       
   local pEntry := _queue[(_write := (_write % QUEUE_SIZE ) + 1)]

   pEntry.receiver  := receiver
   pEntry.param     := parameter
   pEntry.message   := if( symbol   != nil, symbol, _eval)
   pEntry.delay     := if( nDelay   != nil, nDelay, 0)
   pEntry.reinstall := if( lInstall != nil, lInstall, false)
   pEntry.timer     := deadline( pEntry.delay)
   pEntry.pid       := (_pid := (_pid % 4294836225) + 1)

   if _write == _read

      SchedKill()

      alert( "scheduler queue overflow")

      // if this happens, you should increase
      // QUEUE_SIZE or you have a flaw in your application that produces
      // exploding entrys in the process queue. i.e.
      //
      // SchedAdd( {||someFunc()})
      // SchedAdd( {||myFuny()})
      //  .
      //  .
      //  .
      // someFunc() ... SchedAdd({||myFunc()})
      //  .                           /
      //  .   ________/______________/
      //  .  /        \
      //  . /
      // myFunc() ... SchedAdd({||myFunc()})
      //    \                        /
      //     \________/_____________/
      //              \
      //
      // this would bomb the process queue immediately !
      //

   end if

return( _pid)

/* Procedure Ŀ
           Name: SchedAddTimer()                                            
    Description: this should be used to a more long distance running        
                 processes, i.e. database backup, since it uses his own     
                 more convenience process, to make a look three times a     
                 minute if a timer-base process should be executed.         
                 Note: a timer-base process becomes a process of its own    
                       as it is added to the process queue automatic.       
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:04:05pm            Time updated: 05:04:05pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: receiver  - either a codeblock or a object                 
               : symbol    - for a codeblock, this is always 'eval()',      
                             otherwise a method                             
               : parameter - this single value is transferd to the symbol   
                             at execution. Use an array if you need more    
                             than one parameter.                            
               : cTime     - time as "03:30"                                
       See Also: _SchedCheckTimer()                                         
   */
procedure SchedAddTimer( receiver, symbol, parameter, cTime)
//        
   // add task to the list and ...
   //-------------------------------------
   aadd( _toDo, { receiver, symbol, parameter, date() - 1, cTime})

   // ... install the timer checking routine (but only once !)
   //---------------------------------------------------------
   if _isChecking == false
      _isChecking := true
      SchedAdd( {||_SchedCheckTimer()}, nil, nil, 360, true)
   end if

return

/* Procedure Ŀ
           Name: SchedDelTimer()                                            
    Description: clears all timer entrys                                    
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:46:24pm            Time updated: 05:46:24pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: none                                                       
       See Also:                                                            
   */
procedure SchedDelTimer()
//        
   _ToDo  := {}
   _isChecking := false

   // remove the timer checking routine
   //----------------------------------
   SchedKill( {||_SchedCheckTimer()} )

return

/* Procedure Ŀ
           Name: SchedDo()                                                  
    Description: evaluates a Codeblock for every UNREAD entry. The          
                 receiver, message, parameter, delaytime, current timer     
                 and the PID (process ID) are given to the codeblock as     
                 an array so that the entry can be manipulated !            
                 Note: no execution is done during this action !            
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:12:49pm            Time updated: 05:12:49pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: bBlock                                                     
       See Also:                                                            
   */
procedure SchedDo( bBlock)
//        
   local read  := _read
   local write := _write

   SchedDisable()

   do while read != write
      eval( bBlock, _queue[ read])
      if ++read > QUEUE_SIZE
         read := 1
      end if
   end do

   SchedEnable()

return

/* Procedure Ŀ
           Name: SchedDisable()                                             
    Description: Stops the Scheduler                                        
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:13:08pm            Time updated: 05:13:08pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: lVal - could be either true or false                       
       See Also:                                                            
   */
procedure SchedDisable( lVal)
//        
   default lVal to true

   SchedEnable( ! lVal)

return

/* Procedure Ŀ
           Name: SchedEnable()                                              
    Description: Starts the Scheduler                                       
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:13:18pm            Time updated: 05:13:18pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: lVal - could be either true or false                       
       See Also:                                                            
   */
procedure SchedEnable( lVal)
//        
   default lVal to true

   if lVal
      ontickx( "SchedNext", _nDelay, 1)
   else
      ontickx()
   end if

   _isActive := lVal

return

/* Function Ŀ
           Name: SchedInfo()           Docs:                                
    Description: test the troughput an returns it as a string               
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:14:13pm            Time updated: 05:14:13pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: None                                                       
   Return Value: cResult                                                    
       See Also:                                                            
   */
function SchedInfo()
//       
   static lStart := true
   static nTime, cResult

   if lStart == true
      lStart := false

      ontickx()
      _write := _read := 0
      SchedAdd( {|n|_SchedTest(n)}, _eval, 1820)
      ontickx( "SchedNext", 0)

      nTime := seconds()

      do while nTime > 0
      end do

   else
      nTime := seconds() - nTime

      cResult := "Scheduler Kernel V1.00 (c) Ob-Tec Austria 1994:" + chr( 13) + chr( 10);
              + "   Time for test :" + str( nTime, 5, 2) + " sec" + chr( 13) + chr( 10);
              + "   Max. task/sec :" + str( int(1820/nTime), 5) + chr( 13) + chr( 10)

      nTime := 0
   end if

return( cResult)

/* Procedure Ŀ
           Name: SchedInit()                                                
    Description:                                                            
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:15:15pm            Time updated: 05:15:15pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: nTicks   - number of ticks that the scheduler must wait    
                            before he gets another timeslice.               
                            Use a higher value to increase foreground       
                            performance.                                    
                            Use a lower  value to increase backround        
                            execution.                                      
       See Also:                                                            
   */
procedure SchedInit( nTicks)
//        
   local i

   default nTicks to 10

   _nDelay   := nTicks
   _queue    := array( QUEUE_SIZE)
   _isActive := false

   _eval := #:eval()
   _pid  := 0

   for i := 1 to QUEUE_SIZE
      _queue[ i] := { nil, nil, nil, 0, 0, false, 0 }
   next

return

/* Procedure Ŀ
           Name: SchedInstall()                                             
    Description: this is for convenience only. It adds a process to the     
                 that is added to the queue again immediately after         
                 execution                                                  
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-25-95              
   Time created: 05:18:06pm            Time updated: 09:28:22am            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: receiver  - either a codeblock or a object                 
               : symbol    - for a codeblock, this is always 'eval()',      
                             otherwise a method                             
               : parameter - this single value is transferd to the symbol   
                             at execution. Use an array if you need more    
                             than one parameter.                            
               : nDelay    - numbers of ticks to wait until execution       
   Return Value: PID                                                        
       See Also: SchedAdd()                                                 
   */
function SchedInstall( receiver, symbol, parameter, nDelay)
//       
return( SchedAdd( receiver, symbol, parameter, nDelay, true))

/* Function Ŀ
           Name: SchedIsActive()       Docs:                                
    Description: returns true, if the Scheduler is running                  
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:18:29pm            Time updated: 05:18:29pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: None                                                       
   Return Value: true/false                                                 
       See Also:                                                            
   */
function SchedIsActive()
//       
return( _isActive)

/* Function Ŀ
           Name: ScchedIsPending()     Docs:                                
    Description: This routine scans the UNREAD queue entry for a given      
                 object/method pair or an object entry. To find a codeblock 
                 entry you must store and use the PID, received by the      
                 'SchedAdd()' routine.                                      
                 Note: the Scheduler is disable while searching             
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-25-95              
   Time created: 05:19:18pm            Time updated: 09:28:28am            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: receiver  - either a codeblock or a object                 
               : symbol    - for a codeblock, this is always 'eval()',      
                             otherwise a method                             
   Return Value: lFound                                                     
       See Also:                                                            
   */
function SchedIsPending( receiver, symbol)
//       
   local lFound := false

   SchedDisable()

   if valtype( receiver) == "N"

      lFound := (_QueueScan(;
            {| entry | entry.pid == receiver}) > 0)

   else
     if pcount() == 2
         lFound := (_QueueScan(;
            {| entry | valtype( entry.receiver) == valtyp( receiver) .and. ;
               entry.receiver == receiver .and. entry.message:isEqual( symbol)})) > 0

      elseif pcount() == 1
         lFound := (_QueueScan(;
            {| entry | valtype( entry.receiver) == valtyp( receiver) .and. ;
               entry.receiver == receiver }))
      end if
   end if

   SchedEnable()

return( lFound)

/* Procedure Ŀ
           Name: SchedKill()                                                
    Description: Kills either a process or the hole Scheduler.              
                 Note: a process is remove immediately from the queue.      
                       See 'SchedRemove()' for details.                     
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:20:39pm            Time updated: 05:20:39pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: receiver  - either a codeblock, an object or a PID         
               : symbol    - for a codeblock, this is always 'eval()',      
                             otherwise a method                             
       See Also:                                                            
   */
procedure SchedKill( receiver, symbol)
//        
   if pcount() > 0
      SchedRemove( receiver, symbol)
   else
      SchedDisable()
   end if

return

/* Procedure Ŀ
           Name: SchedNext()                                                
    Description: this is the main execution routine.                        
                 Note: the cursor and the color are saved and restored !    
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:23:37pm            Time updated: 05:23:37pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: None                                                       
       See Also:                                                            
   */
procedure SchedNext()
//        
   local event

   local nRow
   local nCol
   local nColor

   // is the queue empty ?
   //----------------------
   if ! _read == _write

      // is it time to execute an entry ?
      //-----------------------------------
      if timedout( (event := _Queue[ (_read := (_read % QUEUE_SIZE) + 1)]).timer)

         // save at least cursor and color settings ...
         //---------------------------------------------
         nRow   := row()
         nCol   := col()
         nColor := setcolor()

         // execute the thread !
         //-----------------------
         event.receiver:perform( event.message, event.param)

         // check, if the process wants to be reinstall automatic
         //-------------------------------------------------------
         if event.reinstall

            // restart the process timer ...
            //-------------------------------
            event.timer := deadline( event.delay)

            // ... and add the process to the queue
            //---------------------------------------
            _Queue[ (_write := (_write % QUEUE_SIZE ) + 1)] := aclone( event)
         end if

         setpos( nRow, nCol)
         setcolor( nColor)

      // .. else append it to the process queue !
      ///-----------------------------------------
      else
         _Queue[ (_write := (_write % QUEUE_SIZE ) + 1)] := aclone( event)
      end if
   end if

return

/* Function Ŀ
           Name: SchedRemove()         Docs:                                
    Description: Kills a process or every occurences of a receiver. This    
                 routine scans all UNREAD entries in the process queue and  
                 tests for either a matching object/method pair or only for 
                 a matching object. You can also use the PID from           
                 'SchedAdd()' to identify a thread.                         
                 Note: the process is killed immediately. This is           
                       accomplished by using a empty codeblock instead of   
                       the original receiver and the reference is lost !    
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:29:31pm            Time updated: 05:29:31pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: receiver  - either a codeblock, an object or an PID        
               : symbol    - for a codeblock, this is always 'eval()',      
                             otherwise a method                             
   Return Value: lRemoved                                                   
       See Also:                                                            
   */
function SchedRemove( receiver, symbol)
//       
   local nPos     := 0
   local lRemoved := false

   SchedDisable()

   if valtype( receiver) == "N"

      do while (nPos := _QueueScan(;
            {| entry | entry.pid == receiver})) > 0

         _queue[ nPos].receiver := {|| nil }
         _queue[ nPos].message  := _eval

         lRemoved := true
      end do

   else

      if ! empty( symbol)

         do while (nPos := _QueueScan(;
            {| entry | valtype( entry.receiver) == valtyp( receiver) .and. ;
               entry.receiver == receiver .and. entry.message:isEqual( symbol)})) > 0

            _queue[ nPos].receiver := {|| nil }
            _queue[ nPos].message  := _eval

            lRemoved := true
         end do

      else

         do while (nPos := _QueueScan(;
            {| entry | valtype( entry.receiver) == valtyp( receiver) .and. ;
               entry.receiver == receiver})) > 0

            _queue[ nPos].receiver := {|| nil }
            _queue[ nPos].message  := _eval

            lRemoved := true
         end do
      end if
   end if

   SchedEnable()

return( lRemoved)

/* Function Ŀ
           Name: _SchedTest()          Docs:                                
    Description: routine is called to determine system performance.         
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:34:53pm            Time updated: 05:34:53pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: nCount                                                     
   Return Value: PID                                                        
       See Also:                                                            
   */
static procedure _SchedTest( nCount)
//               
   if nCount-- > 0
      SchedAdd( {|n|_SchedTest(n)}, _eval, nCount, 0)
   else
      SchedAdd( {||SchedInfo()})
   end if

return

/* Procedure Ŀ
           Name: _SchedCheckTimer()                                         
    Description: if a timer based process is waiting, this routine          
                 checks if it's time to execute it .                        
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:35:40pm            Time updated: 05:35:40pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
     Parameters: None                                                       
       See Also:                                                            
   */
static procedure _SchedCheckTimer()
//               
   local i, entry

   // ... else check the list
   //---------------------------
   for i := 1 to len( _toDo)

      entry := _toDo[ i]

      if left( entry.time, 5) == left( time(), 5) .and. entry.date != date()
         entry.date := date()

         // add the process to the queue ...
         //-----------------------------------
         SchedAdd( entry.receiver, entry.message, entry.param)
      end if

   next

return

/* Function Ŀ
           Name: _QueueScan()          Docs:                                
    Description: evaluates a codeblock for every UNREAD entry in the        
                 process queue until the codeblock evaluates to true        
         Author: M.Rinninger                                                
   Date created: 02-24-95              Date updated: 02-24-95              
   Time created: 05:37:12pm            Time updated: 05:37:12pm            
      Copyright: Ob-Tec Austria                                             
  Ĵ
      Arguments: bBlock                                                     
   Return Value: 0                                                          
       See Also:                                                            
   */
static function _QueueScan( bBlock)
//              
   local read  := _read
   local write := _write

   do while read != write
      if eval( bBlock, _queue[ read])
         return( read)
      end if
      if ++read > QUEUE_SIZE
         read := 1
      end if
   end do

return( 0)
